package com.google.webp;

import android.media.MediaCodec;
import android.media.MediaCodecInfo;
import android.media.MediaCodecList;
import android.media.MediaFormat;
import android.util.Log;
import android.view.Surface;

import com.google.webp.libwebp;

import java.nio.ByteBuffer;
import java.util.Arrays;

/**
 * WebP (lossy/VP8) MediaCodec decoder wrapper.
 * 
 * <p>
 * Rendering to a {@code Surface}:
 * <pre>
 * byte[] webp = ...;
 * MediaCodecWebpDecoder decoder = new MediaCodecWebpDecoder(new Surface(...));
 * MediaCodecWebpDecoder.OutputBuffer image = decoder.decode(webp);
 * decoder.releaseOutputBuffer(true);
 * </pre>
 * <p>
 * Retrieving the decoded image in a byte buffer:
 * <pre>
 * byte[] webp = ...;
 * MediaCodecWebpDecoder decoder = new MediaCodecWebpDecoder(new Surface(...));
 * MediaCodecWebpDecoder.OutputBuffer image = decoder.decode(webp);
 * byte[] array = new byte[image.info.size];
 * array.buffer.rewind();
 * array.buffer.get(array, 0, array.length);
 * decoder.releaseOutputBuffer(false);
 * </pre>
 *
 * @see <a href="http://developer.android.com/reference/android/media/MediaCodec.html">MediaCodec</a>
 */
public class MediaCodecWebpDecoder {
    private static final String TAG = "MediaCodecWebpDecoder";
    private static final boolean VERBOSE = true;
    private static boolean FLUSH_DECODER = true;
    private static final String VP8_MIME_TYPE = "video/x-vnd.on2.vp8";
    private static final long DEQUEUE_INPUT_TIMEOUT = 500000;
    private static final long DEQUEUE_OUTPUT_TIMEOUT = 500000;

    static {
        try {
            System.loadLibrary("webp");
        } catch (UnsatisfiedLinkError e) {
            Log.e(TAG, e.getMessage());
        }
    }

    public static class CodecUnavailableException extends Exception {
        public CodecUnavailableException(String message) { super(message); }
    }

    /**
     * Wrapper class for the ByteBuffer, BufferInfo, MediaFormat returned by MediaCodec.
     */
    public static class OutputBuffer {
        public ByteBuffer buffer;
        public MediaCodec.BufferInfo info;
        public MediaFormat format;
        public int index;

        public OutputBuffer(ByteBuffer b, MediaCodec.BufferInfo i, int idx, MediaFormat fmt) {
            buffer = b;
            info = i;
            index = idx;
            format = fmt;
        }
    }

    /**
     * Creates a MediaCodec based WebP decoder.
     * @param surface Surface to use for rendering
     * @param useSoftwareDecoder if true forces the use of a software decoder
     * @throws CodecUnavailableException if the requested VP8 decoder is unavailable
     * @see android.media.MediaCodec
     */
    public MediaCodecWebpDecoder(Surface surface, boolean useSoftwareDecoder)
            throws CodecUnavailableException {
        init(surface, useSoftwareDecoder);
    }

    /**
     * Creates a MediaCodec based WebP decoder without an associated Surface.
     */
    public MediaCodecWebpDecoder(boolean useSoftwareDecoder) throws CodecUnavailableException {
        init(null, useSoftwareDecoder);
    }

    /**
     * Sets the Surface to be used with the codec.
     *
     * Uses the current width and height of the decoder. The
     * <a href="http://developer.android.com/reference/android/media/MediaCodec.html#configure(android.media.MediaFormat, android.view.Surface, android.media.MediaCrypto, int)">MediaCodec#configure</a>
     * is deferred until the next {@link com.google.webp.MediaCodecWebpDecoder#decode}.
     */
    public void reset(Surface s) {
        reset(s, mWidth, mHeight);
    }

    public void close() {
        if (mDecoder != null) {
            mDecoder.stop();
            mDecoder.release();
            mDecoder = null;
        }
    }

    public int getWidth() { return mWidth; }
    public int getHeight() { return mHeight; }
    /**
     * Returns the configured surface.
     */
    public Surface getSurface() { return mSurface; }
    /**
     * Returns the associated MediaCodec decoder.
     */
    public MediaCodec getMediaCodec() { return mDecoder; }

    /**
     * Releases the buffer returned by decode().
     * @param render if set to true renders the buffer to the configured Surface
     * @see <a href="http://developer.android.com/reference/android/media/MediaCodec.html#releaseOutputBuffer(int, boolean)">android.media.MediaCodec.releaseOutputBuffer</a>
     */
    public void releaseOutputBuffer(boolean render) {
        if (mCurrentOutput == null) return;
        try {
            mDecoder.releaseOutputBuffer(mCurrentOutput.index, render);
            if (FLUSH_DECODER) mDecoder.flush();
        } catch (IllegalStateException ise) {
            Log.e(TAG, "error releasing output buffer, IllegalStateException: " + ise);
        }
        mCurrentOutput = null;
    }

    /**
     * Decodes a WebP image using MediaCodec.
     *
     * @param webp byte array containing a lossy (VP8) WebP image
     * @return {@code OutputBuffer} referencing the decoded image on success, {@code null} otherwise
     * @throws IllegalArgumentException if {@code webp} is not a valid WebP file
     * @see OutputBuffer
     */
    public OutputBuffer decode(byte[] webp) throws IllegalArgumentException {
        getDimensions(webp);

        if (!canUseMediaCodec(webp)) {
            throw new IllegalArgumentException("incompatible webp data");
        }

        if (mCurrentOutput != null) {
            releaseOutputBuffer(false);
        }

        Timer t = new Timer();
        if (mNeedsConfigure) {
            if (!reconfigureVp8Decoder()) {
                return null;
            }
        }

        if (VERBOSE) {
            Log.i(TAG, "decoder start time: " + t.elapsedMs() + "ms");
        }

        MediaCodec.BufferInfo bufferInfo = new MediaCodec.BufferInfo();
        try {
            t.reset();
            int inputBufferIndex = mDecoder.dequeueInputBuffer(DEQUEUE_INPUT_TIMEOUT);
            if (inputBufferIndex < 0) {
                Log.e(TAG, "Unable to dequeue input buffer!");
                mNeedsConfigure = true;
                return null;
            }
            if (VERBOSE) Log.i(TAG, "dequeueInputBuffer: " + t.elapsedMs() + "ms");

            final int vp8_data_offset = 20;
            final int size = webp.length - vp8_data_offset;
            final ByteBuffer inputBuf = mInputBuffers[inputBufferIndex];
            inputBuf.clear();
            inputBuf.put(webp, vp8_data_offset, size);
            inputBuf.rewind();

            t.reset();
            mDecoder.queueInputBuffer(
                    inputBufferIndex,
                    0,          // offset
                    size,
                    mFrames++,  // pts
                    0);         // flags
            if (FLUSH_DECODER) {
                flushDecoder();
            }

            do {
                int result = mDecoder.dequeueOutputBuffer(bufferInfo, DEQUEUE_OUTPUT_TIMEOUT);
                if (result >= 0) {
                    if (VERBOSE) {
                        Log.i(TAG, "output buffer returned, size: " + bufferInfo.size
                                + " elapsed : " + t.elapsedMs() + "ms");
                    }

                    if (bufferInfo.size > 0) {
                        mCurrentOutput = new OutputBuffer(mOutputBuffers[result],
                                                          bufferInfo, result, mOutputFormat);
                        return mCurrentOutput;
                    }
                    return null;
                } else if (result == MediaCodec.INFO_OUTPUT_BUFFERS_CHANGED) {
                    mOutputBuffers = mDecoder.getOutputBuffers();
                } else if (result == MediaCodec.INFO_OUTPUT_FORMAT_CHANGED) {
                    mOutputFormat = mDecoder.getOutputFormat();
                    if (VERBOSE) Log.d(TAG, "output format changed: " + mOutputFormat);
                } else if (result == MediaCodec.INFO_TRY_AGAIN_LATER) {
                    if (VERBOSE) Log.d(TAG, "no output buffer");
                    break;
                } else {
                    Log.e(TAG, "unhandled dequeueOutputBuffer result: " + result);
                    break;
                }
            } while (true);
        } catch (IllegalStateException ise) {
            Log.e(TAG, "Failed to decode webp image");
            mNeedsConfigure = true;
        }

        return null;
    }

    /**
     * Simple nanosecond based timer.
     */
    private static class Timer {
        private static final float TIMESCALE = 1e9f;

        public static float toMs(long time) {
            return time / TIMESCALE * 1000;
        }

        public Timer() { reset(); }

        public void reset() {
            startTime = getTime();
        }

        public long elapsed() {
            return getTime() - startTime;
        }

        public float elapsedMs() {
            return toMs(elapsed());
        }

        // ---------------------------------------------------------------------
        private static long getTime() {
            return System.nanoTime();
        }

        private long startTime;
    }

    private void reset(Surface s, int width, int height) {
        mSurface = s;
        mWidth = width;
        mHeight = height;
        mNeedsConfigure = true;
    }

    private void init(Surface s, boolean useSoftwareDecoder) throws CodecUnavailableException {
        reset(s, 0, 0);
        mForceSoftwareCodec = useSoftwareDecoder;
        mDecoder = createVp8Decoder();
        if (mDecoder == null) {
            throw new CodecUnavailableException(
                    "Unable to create VP8 decoder instance");
        }
    }

    private boolean canUseMediaCodec(byte[] webp) {
        final byte[] RIFF = { 'R', 'I', 'F', 'F' };
        final byte[] WEBPVP8 = { 'W', 'E', 'B', 'P', 'V', 'P', '8', ' ' };
        // 'RIFFxxxxWEBPVP8 xxxx<vp8_10_byte_frame_header>'
        return webp.length >= 30
               && Arrays.equals(Arrays.copyOfRange(webp, 0, 4), RIFF)
               && Arrays.equals(Arrays.copyOfRange(webp, 8, 16), WEBPVP8);
    }

    private void getDimensions(byte[] webp) throws IllegalArgumentException {
        int[] width = new int[1];
        int[] height = new int[1];
        if (libwebp.WebPGetInfo(webp, webp.length, width, height) == 0) {
            throw new IllegalArgumentException("invalid webp data");
        }

        if (width[0] != mWidth || height[0] != mHeight) {
            if (VERBOSE) {
                Log.d(TAG, String.format("size change: %dx%d -> %dx%d",
                      mWidth, mHeight, width[0], height[0]));
            }
            // The software decoder is tolerant of size change, but the hardware decoders are not:
            // force a configure().
            reset(mSurface, width[0], height[0]);
        }
    }

    private MediaCodec createVp8Decoder() {
        final String decoder_name = findVp8Decoder();
        if (decoder_name == null) return null;
        return MediaCodec.createByCodecName(decoder_name);
    }

    private boolean reconfigureVp8Decoder() {
        final MediaFormat format = MediaFormat.createVideoFormat(VP8_MIME_TYPE, mWidth, mHeight);
        try {
            mDecoder.stop();
        } catch (IllegalStateException ise) {
            // This may occur after a failed decode or initial configure. reset()
            // is available for the former in API >= 21, for the rest the state
            // would need to be tracked.
        }

        try {
            mDecoder.configure(format,
                    mSurface,
                    null,      // crypto
                    0);        // flags
            mDecoder.start();
            mInputBuffers = mDecoder.getInputBuffers();
            mOutputBuffers = mDecoder.getOutputBuffers();
            mNeedsConfigure = false;
        } catch (IllegalStateException ise) {
            Log.e(TAG, "Error configuring decoder! IllegalStateException: " + ise);
            return false;
        }
        return true;
    }

    private void flushDecoder() {
        final int inputBufferIndex = mDecoder.dequeueInputBuffer(DEQUEUE_INPUT_TIMEOUT);
        if (inputBufferIndex >= 0) {
            if (VERBOSE) Log.d(TAG, "Sending empty frame");
            mInputBuffers[inputBufferIndex].clear();
            mInputBuffers[inputBufferIndex].rewind();
            mDecoder.queueInputBuffer(inputBufferIndex,
                    0,   // offset
                    0,   // length
                    -1,  // pts
                    MediaCodec.BUFFER_FLAG_END_OF_STREAM);
        } else {
            Log.e(TAG, "Unable to get a buffer to flush the decoder");
        }
    }

    private boolean codecSupportsVp8Decode(MediaCodecInfo info) {
        if (info.isEncoder()) return false;

        final String[] supported_types = info.getSupportedTypes();
        for (int i = 0; i < supported_types.length; ++i) {
            if (supported_types[i].equals(VP8_MIME_TYPE)) return true;
        }
        return false;
    }

    /**
     * Returns true if the decoder supports VP8 decoding of the desired type.
     * Whether the codec is hardware based or not depends on how this instance
     * was constructed.
     * Note this is an approximation as other vendors could ship their own
     * software backed decoder. The check is similar to OMXClient.cpp:IsSoftwareComponent().
     */
    private boolean isVp8Decoder(MediaCodecInfo info) {
        final String name = info.getName();

        final boolean ok = codecSupportsVp8Decode(info);
        if (mForceSoftwareCodec) {
            return ok && name.startsWith("OMX.google");
        }
        return ok && !name.startsWith("OMX.google");
    }

    private boolean isBlacklisted(String name) {
        final String[] blacklist = {
            // "OMX.qcom.",
        };
        for (String codec : blacklist) {
            if (name.startsWith(codec)) return true;
        }
        return false;
    }

    private String findVp8Decoder() {
        for (int i = 0; i < MediaCodecList.getCodecCount(); ++i) {
            MediaCodecInfo info = MediaCodecList.getCodecInfoAt(i);
            if (isVp8Decoder(info)) {
                final String name = info.getName();
                if (isBlacklisted(name)) {
                    Log.v(TAG, "Skipping blacklisted decoder: " + name);
                    continue;
                }
                Log.v(TAG, "Using VP8 decoder: '" + name + "'");
                return info.getName();
            }
        }
        return null;
    }

    // -----------------------------------------------------------------------
    private boolean mForceSoftwareCodec;
    private int mWidth, mHeight;
    private int mFrames;
    private Surface mSurface;
    private MediaCodec mDecoder;
    private ByteBuffer[] mInputBuffers;
    private ByteBuffer[] mOutputBuffers;
    private MediaFormat mOutputFormat;
    private boolean mNeedsConfigure;
    private OutputBuffer mCurrentOutput;
}
